(* +------------------------------------------------------------------------+
   |                        Calcul d'enveloppe convexe                      |
   +------------------------------------------------------------------------+ *)

                             (* +---------------+
                                |  Utilitaires  |
                                +---------------+ *)

#open "num";;
#open "graphics";;
open_graph "";;

let zero = num_of_int(0);;
let deux = num_of_int(2);;

(* pretty-printer *)
let print_num(x) = format__print_string(string_of_num(x));;
install_printer "print_num";;

(* compare deux points *)
let gal (a,b) (c,d) = (a =/ c) && (b =/ d);;

(* tire un tableau de points distincts au hasard *)
let random_points n x0 y0 x1 y1 =
  let p = make_vect n (zero,zero) in
  let i = ref(0)                  in
  while !i < n do
    let a = num_of_int(x0 + random__int(x1-x0)) in
    let b = num_of_int(y0 + random__int(y1-y0)) in
    let m = (a,b)                               in
    let j = ref(0)                              in
    while (!j < !i) && not(gal m p.(!j)) do j := !j + 1 done;
    if !j = !i then begin p.(!i) <- m; i := !i + 1 end
  done;
  p
;;

(* arrondi entier *)
let round(x) = int_of_num(round_num x);;

(* affiche des points *)
let plot_points p =
  set_color black;;
  for i=0 to vect_length(p)-1 do
    let (a,b) = p.(i) in
    fill_circle (round a) (round b) 2
  done
;;

(* affiche une ligne brise *)
let plot_lines p =
  set_color red;;
  let (a,b) = p.(0) in
  moveto (round a) (round b);
  for i=1 to vect_length(p)-1 do
    let (c,d) = p.(i) in
    lineto (round c) (round d)
  done
;;

(* affiche un cercle *)
let plot_circle (a,b) r2 =
  set_color blue;;
  draw_circle (round a) (round b) (int_of_float(sqrt(float_of_num r2)))
;;


                          (* +---------------------+
                             |  Enveloppe convexe  |
                             +---------------------+ *)

(* det(ab,ac) *)
let det (a1,a2) (b1,b2) (c1,c2) = (b1 -/ a1)*/(c2 -/ a2) -/ (b2 -/ a2)*/(c1 -/ a1);;

(* cherche un point extrmal *)
let ordonne_min(p) =
  let n = vect_length(p) in
  for i=1 to n-1 do
    let (a,b) = p.(0) in
    let (c,d) = p.(i) in
    if (d </ b) or ((d =/ b) && (c </ a)) then begin p.(0) <- (c,d); p.(i) <- (a,b) end
  done
;;

(* cherche un point d'angle polaire minimal *)
let angle_min p i =
  let n = vect_length(p) in
  let a = p.(i-1) in
  for j=i+1 to n-1 do
    let b = p.(i) in
    let c = p.(j) in
    if det a b c </ zero then begin p.(i) <- c; p.(j) <- b end
  done
;;

(* algorithme du paquet cadeau *)
let convexe(p) =
  let p = copy_vect(p)            in
  ordonne_min(p);
  let p = concat_vect p [|p.(0)|] in
  let i = ref(1)                  in
  while
    angle_min p !i;
    not(gal p.(0) p.(!i))
  do
    i := !i+1
  done;
  sub_vect p 0 (!i+1)
;;

(* essai *)    
let p = random_points 50 150 150 350 350;;
let c = convexe p;;
clear_graph();;
plot_points p;;
plot_lines c;;

                        (* +------------------------+
                           |  Tests d'appartenance  |
                           +------------------------+ *)

(* teste si un point est intrieur  l'enveloppe convexe *)
let intrieur m p =
  let n   = vect_length(p) in
  let res = ref(true)      in
  let i   = ref(1)         in
  while !res && (!i < n-1) do
    res := det p.(!i-1) p.(!i) m >/ zero;
    i := !i + 1
  done;
  !res
;;

(* teste si un point est sur la frontire *)
let frontire m p =
  let n   = vect_length(p) in
  let res = ref(false)     in
  let i   = ref(1)         in
  while not(!res) && (!i < n-1) do
    if det p.(!i-1) p.(!i) m =/ zero
    then begin
      let (a,b) = p.(!i-1)
      and (c,d) = p.(!i)
      and (e,f) = m       in
      res := ((e-/a)*/(c-/e) >=/ zero) && ((f-/b)*/(d-/f) >=/ zero)
    end;
    i := !i + 1
  done;
  !res
;;

                          (* +--------------------+
                             |  Cercle englobant  |
                             +--------------------+ *)


(* milieu de deux points *)
let milieu (a,b) (c,d) = ((a +/ c)//deux, (b +/ d)//deux);;

(* distance^2 *)
let dist2 (a,b) (c,d) = let x = a -/ c and y = b -/ d in x*/x +/ y*/y;;

(* centre du cercle circonscrit *)
let centre (a1,a2) (b1,b2) (c1,c2) =
  let u  = a1 -/ b1 and v  = a2 -/ b2 and w  = (a1*/a1 +/ a2*/a2 -/ b1*/b1 -/ b2*/b2)//deux in
  let u' = a1 -/ c1 and v' = a2 -/ c2 and w' = (a1*/a1 +/ a2*/a2 -/ c1*/c1 -/ c2*/c2)//deux in
  let d = u*/v' -/ u'*/v in
  ((w*/v' -/ w'*/v)//d, (u*/w' -/ u'*/w)//d)
;;

(* dit si le triangle a tous ses angles aigus *)
let aigu (a1,a2) (b1,b2) (c1,c2) =
     (det (a1,a2) (b1,b2) (c1,c2) <>/ zero)
  && ((b1 -/ a1)*/(c1 -/ a1) +/ (b2 -/ a2)*/(c2 -/ a2) >=/ zero)
  && ((a1 -/ b1)*/(c1 -/ b1) +/ (a2 -/ b2)*/(c2 -/ b2) >=/ zero)
  && ((a1 -/ c1)*/(b1 -/ c1) +/ (a2 -/ c2)*/(b2 -/ c2) >=/ zero)
;;

(* dit si le cercle (c,r2) entoure les points de p *)
let entoure c r2 p =
  let n   = vect_length(p) in
  let i   = ref 0          in
  let res = ref true       in
  while !res && (!i < n-1) do res := dist2 c p.(!i) <=/ r2; i := !i + 1 done;
  !res
;;

(* cercle entourant *)
let cercle(p) =
  let n  = vect_length(p) in
  let c  = ref p.(0)      in
  let r2 = ref(zero)      in
  let trouv = ref false  in

  let i = ref(0) and j = ref(1) in
  while not(!trouv) && (!i < n-1) do
    c  := milieu p.(!i) p.(!j);
    r2 := dist2 !c p.(!i);
    trouv := entoure !c !r2 p;
    j := !j + 1;
    if !j >= n-1 then begin i := !i + 1; j := !i + 1 end
  done;

  let i = ref(0) and j = ref(1) and k = ref(2) in
  while not(!trouv) && (!i < n-2) do
    if aigu p.(!i) p.(!j) p.(!k) then begin
      c  := centre p.(!i) p.(!j) p.(!k);
      r2 := dist2 !c p.(!i);
      trouv := entoure !c !r2 p;
    end;
    k := !k + 1;
    if !k >= n-1 then begin
      j := !j + 1;
      if !j >= n-2 then begin i := !i + 1; j := !i + 1 end;
      k := !j + 1
    end
  done;

  (!c, !r2)
;;

let essai() =
let p = random_points 50 150 150 350 350 in
let c = convexe p in
let (center,d) = cercle(c) in
clear_graph();
plot_points p;
plot_lines c;
plot_circle center d
;;

